#!/usr/bin/env python3

"""Module containing the dmd_run class and the command line interface."""
import argparse
from typing import Optional
from pathlib import Path
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class DMDRun(BiobbObject):
    """
    | biobb_flexserv DMDRun
    | Wrapper of the Discrete Molecular Dynamics tool from the FlexServ module.
    | Generates protein conformational structures using the Discrete Molecular Dynamics (DMD) method.

    Args:
        input_pdb_path (str): Input PDB file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/flexserv/structure.ca.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_log_path (str): Output log file. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/flexserv/dmd_run_out.log>`_. Accepted formats: log (edam:format_2330), out (edam:format_2330), txt (edam:format_2330), o (edam:format_2330).
        output_crd_path (str): Output ensemble. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/flexserv/dmd_run_out.crd>`_. Accepted formats: crd (edam:format_3878), mdcrd (edam:format_3878), inpcrd (edam:format_3878).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **binary_path** (*str*) - ("dmdgoopt") DMD binary path to be used.
            * **dt** (*float*) - (1e-12) Integration time (s)
            * **temperature** (*int*) - (300) Simulation temperature (K)
            * **frames** (*int*) - (1000) Number of frames in the final ensemble
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.flexserv.dmd_run import dmd_run
            prop = {
                'binary_path': 'dmdgoopt'
            }
            flexserv_run(input_pdb_path='/path/to/dmd_input.pdb',
                         output_log_path='/path/to/dmd_log.log',
                         output_crd_path='/path/to/dmd_ensemble.crd',
                         properties=prop)

    Info:
        * wrapped_software:
            * name: FlexServ Discrete Molecular Dynamics
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pdb_path: str, output_log_path: str,
                 output_crd_path: str, properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pdb_path': input_pdb_path},
            'out': {'output_log_path': output_log_path,
                    'output_crd_path': output_crd_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.binary_path = properties.get('binary_path', 'dmdgoopt')
        # self.dt = properties.get('dt', 1.D-12)
        self.dt = properties.get('dt', 1e-12)
        self.temperature = properties.get('temperature', 300)
        self.frames = properties.get('frames', 1000)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ BDRun module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # Internal file paths
        try:
            # Using rel paths to shorten the amount of characters due to fortran path length limitations
            input_pdb = str(Path(self.stage_io_dict["in"]["input_pdb_path"]).relative_to(Path.cwd()))
            output_crd = str(Path(self.stage_io_dict["out"]["output_crd_path"]).relative_to(Path.cwd()))
            output_log = str(Path(self.stage_io_dict["out"]["output_log_path"]).relative_to(Path.cwd()))
        except ValueError:
            # Container or remote case
            input_pdb = self.stage_io_dict["in"]["input_pdb_path"]
            output_crd = self.stage_io_dict["out"]["output_crd_path"]
            output_log = self.stage_io_dict["out"]["output_log_path"]

        # Config file
        instructions_file = str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("dmd.in"))
        with open(instructions_file, 'w') as dmdin:

            dmdin.write("&INPUT\n")
            dmdin.write("  FILE9='{}',\n".format(input_pdb))
            dmdin.write("  FILE21='{}',\n".format(output_crd))
            dmdin.write("  TSNAP={},\n".format(self.dt))
            dmdin.write("  NBLOC={},\n".format(self.frames))
            dmdin.write("  TEMP={},\n".format(self.temperature))
            dmdin.write("  RCUTGO=8,\n")
            dmdin.write("  RCA=0.5,\n")
            dmdin.write("  SIGMA=0.05,\n")
            dmdin.write("  SIGMAGO=0.1,\n")
            dmdin.write("  KKK=2839\n")
            dmdin.write("&END\n")

        # Command line
        # dmdgoopt < dmd.in > dmd.log
        self.cmd = [self.binary_path,
                    '<', instructions_file,
                    '>', output_log
                    ]

        # Run Biobb block
        self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            self.stage_io_dict.get("unique_dir", "")
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def dmd_run(input_pdb_path: str,
            output_log_path: str, output_crd_path: str,
            properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`DMDRun <flexserv.dmd_run.DMDRun>`flexserv.dmd_run.DMDRun class and
    execute :meth:`launch() <flexserv.dmd_run.DMDRun.launch>` method"""

    return DMDRun(input_pdb_path=input_pdb_path,
                  output_log_path=output_log_path,
                  output_crd_path=output_crd_path,
                  properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Generates protein conformational structures using the Discrete Molecular Dynamics method.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pdb_path', required=True, help='Input PDB file. Accepted formats: pdb.')
    required_args.add_argument('--output_log_path', required=True, help='Output log file. Accepted formats: log, out, txt.')
    required_args.add_argument('--output_crd_path', required=True, help='Output ensemble file. Accepted formats: crd, mdcrd, inpcrd.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    dmd_run(input_pdb_path=args.input_pdb_path,
            output_log_path=args.output_log_path,
            output_crd_path=args.output_crd_path,
            properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the PCZhinges class and the command line interface."""
import argparse
from typing import Optional
import shutil
import json
import re
from pathlib import PurePath
from biobb_common.tools import file_utils as fu
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PCZhinges(BiobbObject):
    """
    | biobb_flexserv PCZhinges
    | Compute possible hinge regions (residues around which large protein movements are organized) of a molecule from a compressed PCZ file.
    | Wrapper of the pczdump tool from the PCAsuite FlexServ module.

    Args:
        input_pcz_path (str): Input compressed trajectory file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        output_json_path (str): Output hinge regions x PCA mode file. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/hinges.json>`_. Accepted formats: json (edam:format_3464).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **binary_path** (*str*) - ("pczdump") pczdump binary path to be used.
            * **eigenvector** (*int*) - (0) PCA mode (eigenvector) from which to extract bfactor values per residue (0 means average over all modes).
            * **method** (*str*) - ("Dynamic_domain") Method to compute the hinge regions (Options: Bfactor_slope, Force_constant, Dynamic_domain)
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.pcasuite.pcz_hinges import pcz_hinges
            prop = {
                'eigenvector': 1,
                'pdb': True
            }
            pcz_hinges( input_pcz_path='/path/to/pcazip_input.pcz',
                    output_json_path='/path/to/hinges.json',
                    properties=prop)

    Info:
        * wrapped_software:
            * name: FlexServ PCAsuite
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pcz_path: str, output_json_path: str,
                 properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pcz_path': input_pcz_path},
            'out': {'output_json_path': output_json_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.binary_path = properties.get('binary_path', 'pczdump')
        self.eigenvector = properties.get('eigenvector', 1)
        self.method = properties.get('method', "Bfactor_slope")

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def parse_output(self, output_file):
        """ Parses FlexServ hinges methods output file report """

        method = ''
        if self.method == "Bfactor_slope":
            method = "#### Distance variation method"
        elif self.method == "Force_constant":
            method = "#### Force constant"
        elif self.method == "Dynamic_domain":
            method = "#### Lavery method"
        else:
            print("Method not recognised ({}), please check it and try again. ".format(self.method))

        start = False
        out_data = ''
        with open(output_file, 'r') as file:
            for line in file:
                if method in line:
                    start = True
                elif "####" in line:
                    start = False
                if start:
                    out_data += line

        dict_out = {}
        dict_out["method"] = self.method
        if self.method == "Force_constant":
            dict_out["values_per_residue"] = []
            for line in out_data.split("\n"):
                if line and "#" not in line:
                    dict_out["values_per_residue"].append(float(line.strip()))
                if "possible hinge" in line:  # Peak constant (possible hinge): residue 64 (16.740)
                    residue = int(line.split(' ')[6])
                    dict_out["hinge_residues"] = residue
        elif self.method == "Bfactor_slope":
            dict_out["hinge_residues"] = []
            for line in out_data.split("\n"):
                if "Window" in line:  # Window 28: residue  54 seems a downhill hinge point
                    residue = int(re.split(r'\s+', line)[3])
                    dict_out["hinge_residues"].append(residue)
                if "Consensus" in line:  # Consensus Downhill hinge point :  23.7 (  64.965)
                    hinge_point = float(line.split(':')[1].split('(')[0])
                    dict_out["consensus_hinge"] = hinge_point
        elif self.method == "Dynamic_domain":
            start = 0
            dict_out["clusters"] = []
            for line in out_data.split("\n"):
                if "threshold" not in line and "nClusters" in line:  # nClusters: 2
                    nclusters = int(line.split(':')[1])
                    dict_out["nClusters"] = nclusters
                if "Threshold" in line:  # *** Threshold defined: 0.300000
                    threshold = float(line.split(':')[1])
                    dict_out["threshold"] = threshold
                if "Min. drij" in line:  # *** Min. drij: 0.000322
                    minValue = float(line.split(':')[1])
                    dict_out["minValue"] = minValue
                if "Max. drij" in line:  # *** Max. drij: 6.385425
                    maxValue = float(line.split(':')[1])
                    dict_out["maxValue"] = maxValue
                if "threshold" in line:  # nClusters: 2 threshold: 3.192873
                    final_threshold = float(line.split(':')[2])
                    dict_out["final_threshold"] = final_threshold
                if "Cluster" in line and "elements" in line:  # Cluster 0 (74 elements)
                    clusterLine = line.split()
                    clusterNum = int(clusterLine[1])
                    clusterElems = int(clusterLine[2].replace('(', ''))
                    cluster = {"clusterNum": clusterNum, "clusterElems": clusterElems}
                    dict_out["clusters"].append(cluster)
                    start = start + 1
                if start and "[" in line:
                    # dict_out["clusters"][start-1]["residues"] = list(map(int,list(line.replace(", ]", "").replace("  [","").split(', '))))
                    dict_out["clusters"][start-1]["residues"] = eval(line)
                # Interacting regions: 13 14 30 31 69 70 84 85 112 113 114 115 116 166 167 199 200
                if "Interacting regions" in line:
                    nums = line.split(':')[1]
                    dict_out["interacting_regions"] = list(map(int, nums.split()))
                # Hinge residues: 13 14 30 31 69 70 84 85 112 113 114 115 116 166 167 199 200
                if "Hinge residues" in line:
                    nums = line.split(':')[1]
                    dict_out["hinge_residues"] = list(map(int, nums.split()))

        return dict_out

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ pcz_hinges module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        # self.stage_files()

        # Internal file paths
        # try:
        #     # Using rel paths to shorten the amount of characters due to fortran path length limitations
        #     input_pcz = str(Path(self.stage_io_dict["in"]["input_pcz_path"]).relative_to(Path.cwd()))
        #     output_json = str(Path(self.stage_io_dict["out"]["output_json_path"]).relative_to(Path.cwd()))
        # except ValueError:
        #     # Container or remote case
        #     input_pcz = self.stage_io_dict["in"]["input_pcz_path"]
        #     output_json = self.stage_io_dict["out"]["output_json_path"]

        # Manually creating a Sandbox to avoid issues with input parameters buffer overflow:
        #   Long strings defining a file path makes Fortran or C compiled programs crash if the string
        #   declared is shorter than the input parameter path (string) length.
        #   Generating a temporary folder and working inside this folder (sandbox) fixes this problem.
        #   The problem was found in Galaxy executions, launching Singularity containers (May 2023).

        # Creating temporary folder
        self.tmp_folder = fu.create_unique_dir()
        fu.log('Creating %s temporary folder' % self.tmp_folder, self.out_log)

        shutil.copy2(self.io_dict["in"]["input_pcz_path"], self.tmp_folder)

        # Temporary output
        # temp_out = str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("output.dat"))
        temp_out = "output.dat"
        temp_log = "output.log"
        temp_json = "output.json"

        # Command line (1: dat file)
        # pczdump -i structure.ca.std.pcz --fluc=1 -o bfactor_1.dat
        # self.cmd = [self.binary_path,
        #             "-i", input_pcz,
        #             "-o", temp_out,
        #             "-t", "0.3",
        #             "--hinge={}".format(self.eigenvector),
        #             ">&", "pcz_dump.hinges.log"
        #             ]

        self.cmd = ['cd', self.tmp_folder, ';',
                    self.binary_path,
                    '-i', PurePath(self.io_dict["in"]["input_pcz_path"]).name,
                    '-o', temp_out,
                    "-t", "0.3",
                    "--hinge={}".format(self.eigenvector),
                    ">&", temp_log
                    ]

        # Run Biobb block
        self.run_biobb()

        # Parsing output file and extracting results for the given method
        dict_out = self.parse_output(PurePath(self.tmp_folder).joinpath(temp_out))

        with open(PurePath(self.tmp_folder).joinpath(temp_json), 'w') as out_file:
            out_file.write(json.dumps(dict_out, indent=4))

        # Copy outputs from temporary folder to output path
        shutil.copy2(PurePath(self.tmp_folder).joinpath(temp_json), PurePath(self.io_dict["out"]["output_json_path"]))

        # Copy files to host
        # self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            # self.stage_io_dict.get("unique_dir", ""),
            self.tmp_folder
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def pcz_hinges(input_pcz_path: str, output_json_path: str,
               properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`PCZhinges <flexserv.pcasuite.pcz_hinges>`flexserv.pcasuite.PCZhinges class and
    execute :meth:`launch() <flexserv.pcasuite.pcz_hinges.launch>` method"""

    return PCZhinges(input_pcz_path=input_pcz_path,
                     output_json_path=output_json_path,
                     properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Compute possible hinge regions (residues around which large protein movements are organized) of a molecule from a compressed PCZ file.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pcz_path', required=True, help='Input compressed trajectory file. Accepted formats: pcz.')
    required_args.add_argument('--output_json_path', required=True, help='Output hinge regions x PCA mode file. Accepted formats: json.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    pcz_hinges(input_pcz_path=args.input_pcz_path,
               output_json_path=args.output_json_path,
               properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the PCZsimilarity class and the command line interface."""
import argparse
from typing import Optional
import json
import numpy as np
import shutil
from pathlib import PurePath
from biobb_common.tools import file_utils as fu
from math import exp
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PCZsimilarity(BiobbObject):
    """
    | biobb_flexserv PCZsimilarity
    | Compute PCA similarity between two given compressed PCZ files.
    | Wrapper of the pczdump tool from the PCAsuite FlexServ module.

    Args:
        input_pcz_path1 (str): Input compressed trajectory file 1. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        input_pcz_path2 (str): Input compressed trajectory file 2. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        output_json_path (str): Output json file with PCA Similarity results. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/pcz_similarity.json>`_. Accepted formats: json (edam:format_3464).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **amplifying_factor** (*float*) - ("0.0") common displacement (dx) along the different eigenvectors. If 0, the result is the absolute similarity index (dot product).
            * **binary_path** (*str*) - ("pczdump") pczdump binary path to be used.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.pcasuite.pcz_similarity import pcz_similarity

            pcz_similarity( input_pcz_path1='/path/to/pcazip_input1.pcz',
                    input_pcz_path2='/path/to/pcazip_input2.pcz',
                    output_json_path='/path/to/pcz_similarity.json',
                    properties=prop)

    Info:
        * wrapped_software:
            * name: FlexServ PCAsuite
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pcz_path1: str, input_pcz_path2: str,
                 output_json_path: str, properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pcz_path1': input_pcz_path1,
                   'input_pcz_path2': input_pcz_path2},
            'out': {'output_json_path': output_json_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.amplifying_factor = properties.get('amplifying_factor')
        self.binary_path = properties.get('binary_path', 'pczdump')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    # Check two eigenvectors to be compatible for dot product
    # i.e. same number of vectors and values per vector
    def are_compatible(self, eigenvectors_1, eigenvectors_2):
        # Check the number of eigenvectors and the number of values in both eigenvectors to match
        if len(eigenvectors_1) != len(eigenvectors_2):
            print('WARNING: Number of eigenvectors does not match')
            return False
        if len(eigenvectors_1[0]) != len(eigenvectors_2[0]):
            print('WARNING: Number of values in eigenvectors does not match')
            return False
        return True

    # Weighted Cross Product (WCP).
    # Get the weighted cross product between eigenvectors
    # This is meant to compare PCA results for molecular dynamics structural conformations
    # The number of eigenvectors to be compared may be specified. All (0) by default
    # DISCLAIMER: This code has been translated from a perl script signed by Alberto Perez (13/09/04)
    # Exploring the Essential Dynamics of B-DNA; Alberto Perez, Jose Ramon Blas, Manuel Rueda,
    # Jose Maria Lopez-Bes, Xavier de la Cruz and Modesto Orozco. J. Chem. Theory Comput.2005,1.
    def get_similarity_index(self,
                             eigenvalues_1, eigenvectors_1,
                             eigenvalues_2, eigenvectors_2, dx=None):

        # Check the number of eigenvectors and the number of values in both eigenvectors to match
        if not self.are_compatible(eigenvectors_1, eigenvectors_2):
            raise SystemExit('Eigenvectors are not compatible')

        # Find out the total number of eigenvectors
        # Set the number of eigenvectors to be analyzed in case it is not set or it exceeds the total
        eigenvectors_number = min(len(eigenvectors_1), len(eigenvectors_2))

        # Find out the number of atoms in the structure
        # Eigenvectors are atom coordinates and each atom has 3 coordinates (x, y, z)
        if len(eigenvectors_1[0]) % 3 != 0:
            raise SystemExit('Something is wrong with eigenvectors since number of values is not divisor of 3')
        atom_number = int(len(eigenvectors_1[0]) / 3)

        # Amplifying factor: if it is 0 the algorithm is the same as a simple dot product.
        # The value of the ~10th eigenvalue is usually taken.
        if dx is not None:
            amplifying_factor = dx
        else:
            amplifying_factor = eigenvalues_1[eigenvectors_number-1]

        # Get the denominator
        # Find new denominator
        # for ($i=0;$i<$nvec;$i++){
        #     $cte1+=exp(-1/$val_1[$i]*$ampf);
        #     $cte2+=exp(-1/$val_2[$i]*$ampf);
        #     $part1+=exp(-2/$val_1[$i]*$ampf)*exp(-2/$val_1[$i]*$ampf);
        #     $part2+=exp(-2/$val_2[$i]*$ampf)*exp(-2/$val_2[$i]*$ampf);
        # }
        cte1 = part1 = cte2 = part2 = 0
        for eigenvalue in eigenvalues_1:
            cte1 += exp(-1 / eigenvalue * amplifying_factor)
            part1 += exp(-2 / eigenvalue * amplifying_factor) ** 2
        for eigenvalue in eigenvalues_2:
            cte2 += exp(-1 / eigenvalue * amplifying_factor)
            part2 += exp(-2 / eigenvalue * amplifying_factor) ** 2
        denominator = part1 * cte2 * cte2 / cte1 / cte1 + part2 * cte1 * cte1 / cte2 / cte2

        # Get all eigenvector values together
        eigenvector_values_1 = [v for ev in eigenvectors_1 for v in ev]
        eigenvector_values_2 = [v for ev in eigenvectors_2 for v in ev]

        # IDK what it is doing now
        total_summatory = 0
        for i in range(eigenvectors_number):
            for j in range(eigenvectors_number):
                # Array has vectors in increasing order of vap, get last one first
                a = (eigenvectors_number - 1 - i) * atom_number * 3
                b = (eigenvectors_number - i) * atom_number * 3 - 1
                c = (eigenvectors_number - 1 - j) * atom_number * 3
                d = (eigenvectors_number - j) * atom_number * 3 - 1
                temp1 = eigenvector_values_1[a:b]
                temp2 = eigenvector_values_2[c:d]
                if len(temp1) != len(temp2):
                    raise ValueError("Projection of vectors of different size!!")
                # Project the two vectors
                add = 0
                for k, value_1 in enumerate(temp1):
                    value_2 = temp2[k]
                    add += value_1 * value_2
                add = add * exp(-1 / eigenvalues_1[i] * amplifying_factor - 1 / eigenvalues_2[j] * amplifying_factor)
                add2 = add ** 2
                total_summatory += add2

        similarity_index = total_summatory * 2 / denominator
        return similarity_index

    # Weighted Cross Product (WCP).
    # DF implementation of Alberto's formula
    # Exploring the Essential Dynamics of B-DNA; Alberto Perez, Jose Ramon Blas, Manuel Rueda,
    # Jose Maria Lopez-Bes, Xavier de la Cruz and Modesto Orozco. J. Chem. Theory Comput.2005,1.
    def eigenmsip(self, eigenvalues_1, eigenvectors_1,
                  eigenvalues_2, eigenvectors_2,
                  dx=None):

        evals1 = np.array(eigenvalues_1)
        evals2 = np.array(eigenvalues_2)

        evecs1 = np.array(eigenvectors_1)
        evecs2 = np.array(eigenvectors_2)

        n_components = len(eigenvectors_1)

        # Amplifying factor: if it is 0 the algorithm is the same as a simple dot product.
        # The value of the ~10th eigenvalue is usually taken.
        if dx is not None:
            amplifying_factor = dx
        else:
            amplifying_factor = evals1[n_components-1]

        e1 = np.exp(-(amplifying_factor)**2/evals1)
        e2 = np.exp(-(amplifying_factor)**2/evals2)

        e1_2 = np.exp(-2*(amplifying_factor)**2/evals1)
        e2_2 = np.exp(-2*(amplifying_factor)**2/evals2)
        sume1 = np.sum(e1)
        sume2 = np.sum(e2)

        denominator = np.sum((e1_2/sume1**2)**2)+np.sum((e2_2/sume2**2)**2)

        # numerator_df = np.square(np.dot(evecs1, evecs2)*np.outer(e1, e2)/(sume1*sume2))
        # numerator_df = 2 * np.sum(numerator_df)

        val_tmp = 0
        accum_a = 0
        c = sume1*sume2
        for pc in range(0, n_components):
            for pc2 in range(0, n_components):
                eve1 = evecs1[pc]
                eve2 = evecs2[pc2]
                eva1 = evals1[pc]
                eva2 = evals2[pc2]
                a = np.dot(eve1, eve2)
                b = np.exp(-(amplifying_factor)**2/eva1 - (amplifying_factor)**2/eva2)
                val_tmp = val_tmp + ((a * b)/c)**2
                accum_a = accum_a + a

        numerator = 2 * val_tmp

        return numerator/(denominator)

    # Get the dot product matrix of two eigenvectors
    def dot_product(self, eigenvectors_1, eigenvectors_2):
        # Check the number of eigenvectors and the number of values in both eigenvectors to match
        if not self.are_compatible(eigenvectors_1, eigenvectors_2):
            raise SystemExit('Eigenvectors are not compatible')
        # Get the dot product
        dpm = np.dot(eigenvectors_1, np.transpose(eigenvectors_2))
        return dpm

    # Get the dot product matrix of two eigenvectors (squared and normalized).
    # Absolute Similarity Index (Hess 2000, 2002)
    def dot_product_accum(self, eigenvectors_1, eigenvectors_2):
        n_components = len(eigenvectors_1)
        # Get the dot product
        dpm = self.dot_product(eigenvectors_1, eigenvectors_2)

        sso = (dpm * dpm).sum() / n_components
        # sso = (dpm * dpm).sum() / n_components
        return sso

    # Get the subspace overlap
    # Same as before, but with a square root
    def get_subspace_overlap(self, eigenvectors_1, eigenvectors_2):
        # Get the number of eigenvectors
        n_components = len(eigenvectors_1)
        # Get the dot product
        dpm = self.dot_product(eigenvectors_1, eigenvectors_2)

        sso = np.sqrt((dpm * dpm).sum() / n_components)
        # sso = (dpm * dpm).sum() / n_components
        return sso

    # Classic RMSip (Root Mean Square Inner Product), gives the same results as the previous function get_subspace_overlap
    def get_rmsip(self, eigenvectors_1, eigenvectors_2):

        # Get the number of eigenvectors
        n_components = len(eigenvectors_1)

        accum = 0
        for pc in range(0, n_components):
            for pc2 in range(0, n_components):
                dpm = np.dot(eigenvectors_1[pc], eigenvectors_2[pc2])
                val = dpm * dpm
                accum = accum + val

        sso = np.sqrt(accum / n_components)
        # sso = (dpm * dpm).sum() / n_components
        return sso

    # RWSIP, Root Weighted Square Inner Product. Same as before, but weighted using the eigen values.
    # See Edvin Fuglebakk and others, Measuring and comparing structural fluctuation patterns in large protein datasets,
    # Bioinformatics, Volume 28, Issue 19, October 2012, Pages 2431–2440, https://doi.org/10.1093/bioinformatics/bts445
    def get_rwsip(self,
                  eigenvalues_1, eigenvectors_1,
                  eigenvalues_2, eigenvectors_2):

        # Get the number of eigenvectors
        n_components = len(eigenvectors_1)

        accum = 0
        norm = 0
        for pc in range(0, n_components):
            for pc2 in range(0, n_components):
                dpm = np.dot(eigenvectors_1[pc], eigenvectors_2[pc2])
                val = dpm * dpm * eigenvalues_1[pc] * eigenvalues_2[pc2]
                accum = accum + val
            norm = norm + eigenvalues_1[pc] * eigenvalues_2[pc]

        sso = np.sqrt(accum / norm)
        # sso = (dpm * dpm).sum() / n_components
        return sso

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ pcz_similarity module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        # self.stage_files()

        # Internal file paths
        # try:
        #     # Using rel paths to shorten the amount of characters due to fortran path length limitations
        #     input_pcz_1 = str(Path(self.stage_io_dict["in"]["input_pcz_path1"]).relative_to(Path.cwd()))
        #     input_pcz_2 = str(Path(self.stage_io_dict["in"]["input_pcz_path2"]).relative_to(Path.cwd()))
        #     output_json = str(Path(self.stage_io_dict["out"]["output_json_path"]).relative_to(Path.cwd()))
        # except ValueError:
        #     # Container or remote case
        #     output_json = self.stage_io_dict["out"]["output_json_path"]

        # Manually creating a Sandbox to avoid issues with input parameters buffer overflow:
        #   Long strings defining a file path makes Fortran or C compiled programs crash if the string
        #   declared is shorter than the input parameter path (string) length.
        #   Generating a temporary folder and working inside this folder (sandbox) fixes this problem.
        #   The problem was found in Galaxy executions, launching Singularity containers (May 2023).

        # Creating temporary folder
        self.tmp_folder = fu.create_unique_dir()
        fu.log('Creating %s temporary folder' % self.tmp_folder, self.out_log)

        shutil.copy2(self.io_dict["in"]["input_pcz_path1"], self.tmp_folder)
        shutil.copy2(self.io_dict["in"]["input_pcz_path2"], self.tmp_folder)

        # Temporary output
        temp_json = "output.json"
        temp_out_1 = "output1.dat"
        temp_out_2 = "output2.dat"

        # Command line 1
        # pczdump -i structure.ca.std.pcz --evals -o evals.txt
        # self.cmd = [self.binary_path,   # Evals pcz 1
        #             "-i", input_pcz_1,
        #             "-o", temp_out_1,
        #             "--evals", ';',
        #             self.binary_path,   # Evals pcz 2
        #             "-i", input_pcz_2,
        #             "-o", temp_out_2,
        #             "--evals"
        #             ]

        self.cmd = ['cd', self.tmp_folder, ';',
                    self.binary_path,    # Evals pcz 1
                    "-i", PurePath(self.io_dict["in"]["input_pcz_path1"]).name,
                    "-o", temp_out_1,
                    "--evals", ';',
                    self.binary_path,    # Evals pcz 2
                    "-i", PurePath(self.io_dict["in"]["input_pcz_path2"]).name,
                    "-o", temp_out_2,
                    "--evals"
                    ]

        # Run Biobb block 1
        self.run_biobb()

        # Parse output evals
        info_dict = {}
        info_dict['evals_1'] = []
        info_dict['evals_2'] = []

        with open(PurePath(self.tmp_folder).joinpath(temp_out_1), 'r') as file:
            for line in file:
                info = float(line.strip())
                info_dict['evals_1'].append(info)

        with open(PurePath(self.tmp_folder).joinpath(temp_out_2), 'r') as file:
            for line in file:
                info = float(line.strip())
                info_dict['evals_2'].append(info)

        num_evals_1 = len(info_dict['evals_1'])
        num_evals_2 = len(info_dict['evals_2'])
        num_evals_min = min(num_evals_1, num_evals_2)
        info_dict['num_evals_min'] = num_evals_min

        # Command line 2
        # pczdump -i structure.ca.std.pcz --evals -o evals.txt
        self.cmd = []
        self.cmd.append('cd')
        self.cmd.append(self.tmp_folder)
        self.cmd.append(';')
        for pc in (range(1, num_evals_min+1)):
            # Evecs pcz 1
            self.cmd.append(self.binary_path)
            self.cmd.append("-i")
            # self.cmd.append(input_pcz_1)
            self.cmd.append(PurePath(self.io_dict["in"]["input_pcz_path1"]).name)
            # self.cmd.append("-o evecs_1_pc{}".format(pc))
            self.cmd.append("-o")
            # self.cmd.append(str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("evecs_1_pc{}".format(pc))))
            self.cmd.append("evecs_1_pc{}".format(pc))
            self.cmd.append("--evec={}".format(pc))
            self.cmd.append(";")
            # Evals pcz 2
            self.cmd.append(self.binary_path)
            self.cmd.append("-i")
            # self.cmd.append(input_pcz_2)
            self.cmd.append(PurePath(self.io_dict["in"]["input_pcz_path2"]).name)
            # self.cmd.append("-o evecs_2_pc{}".format(pc))
            self.cmd.append("-o")
            # self.cmd.append(str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("evecs_2_pc{}".format(pc))))
            self.cmd.append("evecs_2_pc{}".format(pc))
            self.cmd.append("--evec={}".format(pc))
            self.cmd.append(";")

        # Run Biobb block 2
        self.run_biobb()

        # Parse output evecs
        info_dict['evecs_1'] = {}
        info_dict['evecs_2'] = {}
        eigenvectors_1 = []
        eigenvectors_2 = []
        for pc in (range(1, num_evals_min+1)):
            pc_id = "pc{}".format(pc)
            info_dict['evecs_1'][pc_id] = []
            info_dict['evecs_2'][pc_id] = []
            # with open(str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("evecs_1_pc{}".format(pc))), 'r') as file:
            with open(PurePath(self.tmp_folder).joinpath("evecs_1_pc{}".format(pc)), 'r') as file:
                list_evecs = []
                for line in file:
                    info = line.strip().split(' ')
                    for nums in info:
                        if nums:
                            list_evecs.append(float(nums))
                info_dict['evecs_1'][pc_id] = list_evecs
                eigenvectors_1.append(list_evecs)

            # with open(str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("evecs_2_pc{}".format(pc))), 'r') as file:
            with open(PurePath(self.tmp_folder).joinpath("evecs_2_pc{}".format(pc)), 'r') as file:
                list_evecs = []
                for line in file:
                    info = line.strip().split(' ')
                    for nums in info:
                        if nums:
                            list_evecs.append(float(nums))
                info_dict['evecs_2'][pc_id] = list_evecs
                eigenvectors_2.append(list_evecs)

        # simIndex = self.get_similarity_index(info_dict['evals_1'], eigenvectors_1, info_dict['evals_2'], eigenvectors_2, self.amplifying_factor)
        # info_dict['similarityIndex_WCP2'] = float("{:.3f}".format(simIndex))
        # dotProduct = self.get_subspace_overlap(eigenvectors_1, eigenvectors_2)
        # info_dict['similarityIndex_rmsip2'] = float("{:.3f}".format(dotProduct))
        eigenmsip = self.eigenmsip(info_dict['evals_1'], eigenvectors_1, info_dict['evals_2'], eigenvectors_2, self.amplifying_factor)
        info_dict['similarityIndex_WCP'] = float("{:.3f}".format(eigenmsip))
        rmsip = self.get_rmsip(eigenvectors_1, eigenvectors_2)
        info_dict['similarityIndex_rmsip'] = float("{:.3f}".format(rmsip))
        rwsip = self.get_rwsip(info_dict['evals_1'], eigenvectors_1, info_dict['evals_2'], eigenvectors_2)
        info_dict['similarityIndex_rwsip'] = float("{:.3f}".format(rwsip))
        dotp = self.dot_product_accum(eigenvectors_1, eigenvectors_2)
        info_dict['similarityIndex_dotp'] = float("{:.3f}".format(dotp))

        with open(PurePath(self.tmp_folder).joinpath(temp_json), 'w') as out_file:
            out_file.write(json.dumps(info_dict, indent=4))

        # Copy outputs from temporary folder to output path
        shutil.copy2(PurePath(self.tmp_folder).joinpath(temp_json), PurePath(self.io_dict["out"]["output_json_path"]))

        # remove temporary folder(s)
        self.tmp_files.extend([
            self.tmp_folder
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def pcz_similarity(input_pcz_path1: str, input_pcz_path2: str, output_json_path: str,
                   properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`PCZsimilarity <flexserv.pcasuite.pcz_similarity>`flexserv.pcasuite.PCZsimilarity class and
    execute :meth:`launch() <flexserv.pcasuite.pcz_similarity.launch>` method"""

    return PCZsimilarity(input_pcz_path1=input_pcz_path1,
                         input_pcz_path2=input_pcz_path2,
                         output_json_path=output_json_path,
                         properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Compute PCA Similarity from a given pair of compressed PCZ files.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pcz_path1', required=True, help='Input compressed trajectory file 1. Accepted formats: pcz.')
    required_args.add_argument('--input_pcz_path2', required=True, help='Input compressed trajectory file 2. Accepted formats: pcz.')
    required_args.add_argument('--output_json_path', required=True, help='Output json file with PCA similarity. Accepted formats: json.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    pcz_similarity(input_pcz_path1=args.input_pcz_path1,
                   input_pcz_path2=args.input_pcz_path2,
                   output_json_path=args.output_json_path,
                   properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the PCZbfactor class and the command line interface."""
import argparse
from typing import Optional
import shutil
from pathlib import PurePath
from biobb_common.tools import file_utils as fu
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PCZbfactor(BiobbObject):
    """
    | biobb_flexserv PCZbfactor
    | Extract residue bfactors x PCA mode from a compressed PCZ file.
    | Wrapper of the pczdump tool from the PCAsuite FlexServ module.

    Args:
        input_pcz_path (str): Input compressed trajectory file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        output_dat_path (str): Output Bfactor x residue x PCA mode file. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/bfactors.dat>`_. Accepted formats: dat (edam:format_1637), txt (edam:format_2330), csv (edam:format_3752).
        output_pdb_path (str) (Optional): Output PDB with Bfactor x residue x PCA mode file. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/bfactors.pdb>`_. Accepted formats: pdb (edam:format_1476).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **binary_path** (*str*) - ("pczdump") pczdump binary path to be used.
            * **eigenvector** (*int*) - (0) PCA mode (eigenvector) from which to extract bfactor values per residue (0 means average over all modes).
            * **pdb** (*bool*) - (False) Generate a PDB file with the computed bfactors (to be easily represented with colour scale)
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.pcasuite.pcz_bfactor import pcz_bfactor
            prop = {
                'eigenvector': 1,
                'pdb': True
            }
            pcz_bfactor( input_pcz_path='/path/to/pcazip_input.pcz',
                    output_dat_path='/path/to/bfactors_mode1.dat',
                    output_pdb_path='/path/to/bfactors_mode1.pdb',
                    properties=prop)

    Info:
        * wrapped_software:
            * name: FlexServ PCAsuite
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pcz_path: str, output_dat_path: str,
                 output_pdb_path: str, properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pcz_path': input_pcz_path},
            'out': {'output_dat_path': output_dat_path,
                    'output_pdb_path': output_pdb_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.binary_path = properties.get('binary_path', 'pczdump')
        self.eigenvector = properties.get('eigenvector', 1)
        self.pdb = properties.get('pdb', False)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ pcz_bfactor module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        # self.stage_files()

        # # Internal file paths
        # try:
        #     # Using rel paths to shorten the amount of characters due to fortran path length limitations
        #     input_pcz = str(Path(self.stage_io_dict["in"]["input_pcz_path"]).relative_to(Path.cwd()))
        #     output_pdb = str(Path(self.stage_io_dict["out"]["output_pdb_path"]).relative_to(Path.cwd()))
        #     output_dat = str(Path(self.stage_io_dict["out"]["output_dat_path"]).relative_to(Path.cwd()))
        # except ValueError:
        #     # Container or remote case
        #     input_pcz = self.stage_io_dict["in"]["input_pcz_path"]
        #     output_pdb = self.stage_io_dict["out"]["output_pdb_path"]
        #     output_dat = self.stage_io_dict["out"]["output_dat_path"]

        # Manually creating a Sandbox to avoid issues with input parameters buffer overflow:
        #   Long strings defining a file path makes Fortran or C compiled programs crash if the string
        #   declared is shorter than the input parameter path (string) length.
        #   Generating a temporary folder and working inside this folder (sandbox) fixes this problem.
        #   The problem was found in Galaxy executions, launching Singularity containers (May 2023).

        # Creating temporary folder
        self.tmp_folder = fu.create_unique_dir()
        fu.log('Creating %s temporary folder' % self.tmp_folder, self.out_log)

        shutil.copy2(self.io_dict["in"]["input_pcz_path"], self.tmp_folder)

        # Command line (1: dat file)
        # pczdump -i structure.ca.std.pcz --fluc=1 -o bfactor_1.dat
        # self.cmd = [self.binary_path,
        #             "-i", input_pcz,
        #             "-o", output_dat,
        #             "--bfactor",
        #             "--fluc={}".format(self.eigenvector)
        #             ]

        self.cmd = ['cd', self.tmp_folder, ';',
                    self.binary_path,
                    '-i', PurePath(self.io_dict["in"]["input_pcz_path"]).name,
                    '-o', PurePath(self.io_dict["out"]["output_dat_path"]).name,
                    "--bfactor",
                    "--fluc={}".format(self.eigenvector)
                    ]

        # Run Biobb block
        self.run_biobb()

        if self.pdb:
            # Command line (2: pdb file)
            # pczdump -i structure.ca.std.pcz --fluc=1 --pdb -o bfactor_1.pdb
            # self.cmd = [self.binary_path,
            #             "-i", input_pcz,
            #             "-o", output_pdb,
            #             "--bfactor",
            #             "--fluc={}".format(self.eigenvector),
            #             "--pdb"
            #             ]

            self.cmd = ['cd', self.tmp_folder, ';',
                        self.binary_path,
                        '-i', PurePath(self.io_dict["in"]["input_pcz_path"]).name,
                        '-o', PurePath(self.io_dict["out"]["output_pdb_path"]).name,
                        "--bfactor",
                        "--fluc={}".format(self.eigenvector),
                        "--pdb"
                        ]

            # Run Biobb block
            self.run_biobb()

        # Copy outputs from temporary folder to output path
        shutil.copy2(PurePath(self.tmp_folder).joinpath(PurePath(self.io_dict["out"]["output_dat_path"]).name), PurePath(self.io_dict["out"]["output_dat_path"]))

        if self.pdb:
            shutil.copy2(PurePath(self.tmp_folder).joinpath(PurePath(self.io_dict["out"]["output_pdb_path"]).name), PurePath(self.io_dict["out"]["output_pdb_path"]))

        # Copy files to host
        # self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            # self.stage_io_dict.get("unique_dir", ""),
            self.tmp_folder
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def pcz_bfactor(input_pcz_path: str, output_dat_path: str, output_pdb_path: str,
                properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`PCZbfactor <flexserv.pcasuite.pcz_bfactor>`flexserv.pcasuite.PCZbfactor class and
    execute :meth:`launch() <flexserv.pcasuite.pcz_bfactor.launch>` method"""

    return PCZbfactor(input_pcz_path=input_pcz_path,
                      output_dat_path=output_dat_path,
                      output_pdb_path=output_pdb_path,
                      properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Extract residue bfactors x PCA mode from a compressed PCZ file.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pcz_path', required=True, help='Input compressed trajectory file. Accepted formats: pcz.')
    required_args.add_argument('--output_dat_path', required=True, help='Output Bfactor x residue x PCA mode file. Accepted formats: dat, txt, csv.')
    required_args.add_argument('--output_pdb_path', required=False, help='Output PDB with Bfactor x residue x PCA mode file. Accepted formats: pdb.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    pcz_bfactor(input_pcz_path=args.input_pcz_path,
                output_dat_path=args.output_dat_path,
                output_pdb_path=args.output_pdb_path,
                properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the PCZinfo class and the command line interface."""
import argparse
from typing import Optional
import shutil
import json
from pathlib import PurePath
from biobb_common.tools import file_utils as fu
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PCZinfo(BiobbObject):
    """
    | biobb_flexserv PCZinfo
    | Extract PCA info (variance, Dimensionality) from a compressed PCZ file.
    | Wrapper of the pczdump tool from the PCAsuite FlexServ module.

    Args:
        input_pcz_path (str): Input compressed trajectory file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        output_json_path (str): Output json file with PCA info such as number of components, variance and dimensionality. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/pcz_info.json>`_. Accepted formats: json (edam:format_3464).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **binary_path** (*str*) - ("pczdump") pczdump binary path to be used.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.pcasuite.pcz_info import pcz_info

            pcz_info( input_pcz_path='/path/to/pcazip_input.pcz',
                    output_json_path='/path/to/pcz_info.json')

    Info:
        * wrapped_software:
            * name: FlexServ PCAsuite
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pcz_path: str,
                 output_json_path: str, properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pcz_path': input_pcz_path},
            'out': {'output_json_path': output_json_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.binary_path = properties.get('binary_path', 'pczdump')

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ pcz_info module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        # self.stage_files()

        # Internal file paths
        # try:
        #     # Using rel paths to shorten the amount of characters due to fortran path length limitations
        #     input_pcz = str(Path(self.stage_io_dict["in"]["input_pcz_path"]).relative_to(Path.cwd()))
        #     output_json = str(Path(self.stage_io_dict["out"]["output_json_path"]).relative_to(Path.cwd()))
        # except ValueError:
        #     # Container or remote case
        #     input_pcz = self.stage_io_dict["in"]["input_pcz_path"]
        #     output_json = self.stage_io_dict["out"]["output_json_path"]

        # Manually creating a Sandbox to avoid issues with input parameters buffer overflow:
        #   Long strings defining a file path makes Fortran or C compiled programs crash if the string
        #   declared is shorter than the input parameter path (string) length.
        #   Generating a temporary folder and working inside this folder (sandbox) fixes this problem.
        #   The problem was found in Galaxy executions, launching Singularity containers (May 2023).

        # Creating temporary folder
        self.tmp_folder = fu.create_unique_dir()
        fu.log('Creating %s temporary folder' % self.tmp_folder, self.out_log)

        shutil.copy2(self.io_dict["in"]["input_pcz_path"], self.tmp_folder)

        # Temporary output
        # temp_out_1 = str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("output1.dat"))
        # temp_out_2 = str(Path(self.stage_io_dict.get("unique_dir", "")).joinpath("output2.dat"))
        temp_out_1 = "output1.dat"
        temp_out_2 = "output2.dat"
        temp_json = "output.json"

        # Command line
        # pczdump -i structure.ca.std.pcz --info -o pcz.info
        # self.cmd = [self.binary_path,
        #             "-i", input_pcz,
        #             "-o", temp_out_1,
        #             "--info", ';',
        #             self.binary_path,
        #             "-i", input_pcz,
        #             "-o", temp_out_2,
        #             "--evals"
        #             ]

        self.cmd = ['cd', self.tmp_folder, ';',
                    self.binary_path,
                    "-i", PurePath(self.io_dict["in"]["input_pcz_path"]).name,
                    "-o", temp_out_1,
                    "--info", ';',
                    self.binary_path,
                    "-i", PurePath(self.io_dict["in"]["input_pcz_path"]).name,
                    "-o", temp_out_2,
                    "--evals"
                    ]

        # Run Biobb block
        self.run_biobb()

        # Parse output info
        # Title             : MC generated trajectory
        # Atoms             :       85
        # Vectors           :        4
        # Frames            :     1000
        # Total variance    :  1137.20
        # Explained variance:  1043.32
        # Quality           :    91.74%
        # Dimensionality    :       21
        # RMSd type         : Standard RMSd
        # Have atom names   : True
        info_dict = {}
        with open(PurePath(self.tmp_folder).joinpath(temp_out_1), 'r') as file:
            for line in file:
                info = line.split(':')
                info_dict[info[0].strip().replace(' ', '_')] = info[1].strip()

        # Parse output evals
        # 744.201782
        # 170.061981
        # 89.214905
        # 39.836308
        info_dict['Eigen_Values'] = []
        info_dict['Eigen_Values_dimensionality_vs_total'] = []
        info_dict['Eigen_Values_dimensionality_vs_explained'] = []
        accum_tot = 0
        accum_exp = 0
        with open(PurePath(self.tmp_folder).joinpath(temp_out_2), 'r') as file:
            for line in file:
                eval = float(line.strip())
                eval_var = (eval / float(info_dict['Total_variance']))*100
                accum_tot = accum_tot + eval_var
                eval_dim = (eval / float(info_dict['Explained_variance']))*100
                accum_exp = accum_exp + eval_dim
                info_dict['Eigen_Values'].append(eval)
                info_dict['Eigen_Values_dimensionality_vs_total'].append(accum_tot)
                info_dict['Eigen_Values_dimensionality_vs_explained'].append(accum_exp)

        with open(PurePath(self.tmp_folder).joinpath(temp_json), 'w') as out_file:
            out_file.write(json.dumps(info_dict, indent=4))

        # Copy outputs from temporary folder to output path
        shutil.copy2(PurePath(self.tmp_folder).joinpath(temp_json), PurePath(self.io_dict["out"]["output_json_path"]))

        # Copy files to host
        # self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            # self.stage_io_dict.get("unique_dir", ""),
            self.tmp_folder
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def pcz_info(input_pcz_path: str, output_json_path: str,
             properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`PCZinfo <flexserv.pcasuite.pcz_info>`flexserv.pcasuite.PCZinfo class and
    execute :meth:`launch() <flexserv.pcasuite.pcz_info.launch>` method"""

    return PCZinfo(input_pcz_path=input_pcz_path,
                   output_json_path=output_json_path,
                   properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Extract PCA info from a compressed PCZ file.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pcz_path', required=True, help='Input compressed trajectory file. Accepted formats: pcz.')
    required_args.add_argument('--output_json_path', required=True, help='Output json file with PCA info such as number of components, variance and dimensionality. Accepted formats: json.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    pcz_info(input_pcz_path=args.input_pcz_path,
             output_json_path=args.output_json_path,
             properties=properties)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the PCAzip class and the command line interface."""
import argparse
from typing import Optional
import shutil
from pathlib import PurePath
from biobb_common.tools import file_utils as fu
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.configuration import settings
from biobb_common.tools.file_utils import launchlogger


class PCZzip(BiobbObject):
    """
    | biobb_flexserv PCZzip
    | Wrapper of the pcazip tool from the PCAsuite FlexServ module.
    | Compress Molecular Dynamics (MD) trajectories using Principal Component Analysis (PCA) algorithms.

    Args:
        input_pdb_path (str): Input PDB file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/structure.ca.pdb>`_. Accepted formats: pdb (edam:format_1476).
        input_crd_path (str): Input Trajectory file. File type: input. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/data/pcasuite/traj.crd>`_. Accepted formats: crd (edam:format_3878), mdcrd (edam:format_3878), inpcrd (edam:format_3878).
        output_pcz_path (str): Output compressed trajectory. File type: output. `Sample file <https://github.com/bioexcel/biobb_flexserv/raw/master/biobb_flexserv/test/reference/pcasuite/pcazip.pcz>`_. Accepted formats: pcz (edam:format_3874).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **binary_path** (*str*) - ("pcazip") pcazip binary path to be used.
            * **neigenv** (*int*) - (0) Number of generated eigenvectors
            * **variance** (*int*) - (90) Percentage of variance captured by the final set of eigenvectors
            * **verbose** (*bool*) - (False) Make output verbose
            * **gauss_rmsd** (*bool*) - (False) Use a gaussian RMSd for fitting
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_flexserv.pcasuite.pcz_zip import pcz_zip
            prop = {
                'variance': 90
            }
            pcz_zip( input_pdb_path='/path/to/pcazip_input.pdb',
                    input_crd_path='/path/to/pcazip_input.crd',
                    output_pcz_path='/path/to/pcazip_traj.pcz',
                    properties=prop)

    Info:
        * wrapped_software:
            * name: FlexServ PCAsuite
            * version: >=1.0
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_pdb_path: str, input_crd_path: str,
                 output_pcz_path: str, properties: Optional[dict] = None, **kwargs) -> None:

        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {'input_pdb_path': input_pdb_path,
                   'input_crd_path': input_crd_path},
            'out': {'output_pcz_path': output_pcz_path}
        }

        # Properties specific for BB
        self.properties = properties
        self.binary_path = properties.get('binary_path', 'pcazip')
        self.neigenv = properties.get('neigenv', 0)
        # self.variance = properties.get('variance', 90)
        self.variance = properties.get('variance')
        self.verbose = properties.get('verbose', False)
        self.gauss_rmsd = properties.get('gauss_rmsd', False)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self):
        """Launches the execution of the FlexServ pcazip module."""

        # Setup Biobb
        if self.check_restart():
            return 0
        # self.stage_files()

        # try:
        #     # Using rel paths to shorten the amount of characters due to fortran path length limitations
        #     input_pdb = str(Path(self.stage_io_dict["in"]["input_pdb_path"]).relative_to(Path.cwd()))
        #     input_crd = str(Path(self.stage_io_dict["in"]["input_crd_path"]).relative_to(Path.cwd()))
        #     output_pcz = str(Path(self.stage_io_dict["out"]["output_pcz_path"]).relative_to(Path.cwd()))
        # except ValueError:
        #     # Container or remote case
        #     input_pdb = self.stage_io_dict["in"]["input_pdb_path"]
        #     input_crd = self.stage_io_dict["in"]["input_crd_path"]
        #     output_pcz = self.stage_io_dict["out"]["output_pcz_path"]

        # Manually creating a Sandbox to avoid issues with input parameters buffer overflow:
        #   Long strings defining a file path makes Fortran or C compiled programs crash if the string
        #   declared is shorter than the input parameter path (string) length.
        #   Generating a temporary folder and working inside this folder (sandbox) fixes this problem.
        #   The problem was found in Galaxy executions, launching Singularity containers (May 2023).

        # Creating temporary folder
        self.tmp_folder = fu.create_unique_dir()
        fu.log('Creating %s temporary folder' % self.tmp_folder, self.out_log)

        shutil.copy2(self.io_dict["in"]["input_pdb_path"], self.tmp_folder)
        shutil.copy2(self.io_dict["in"]["input_crd_path"], self.tmp_folder)

        # Command line
        # pcazip -i infile -o outfile -n natoms
        # [-v] [--mask maskfile] [-e nev] [-q qual] [--pdb pdbfile]
        # self.cmd = [self.binary_path,
        #             "-p", input_pdb,
        #             "-i", input_crd,
        #             "-o", output_pcz
        #             ]

        self.cmd = ['cd', self.tmp_folder, ';',
                    self.binary_path,
                    "-p", PurePath(self.io_dict["in"]["input_pdb_path"]).name,
                    "-i", PurePath(self.io_dict["in"]["input_crd_path"]).name,
                    "-o", PurePath(self.io_dict["out"]["output_pcz_path"]).name
                    ]

        if self.verbose:
            self.cmd.append('-v')

        if self.gauss_rmsd:
            self.cmd.append('-g')

        if self.neigenv:
            self.cmd.append('-e')
            self.cmd.append(str(self.neigenv))

        if self.variance:
            self.cmd.append('-q')
            self.cmd.append(str(self.variance))

        # Run Biobb block
        self.run_biobb()

        # Copy outputs from temporary folder to output path
        shutil.copy2(PurePath(self.tmp_folder).joinpath(PurePath(self.io_dict["out"]["output_pcz_path"]).name), PurePath(self.io_dict["out"]["output_pcz_path"]))

        # Copy files to host
        # self.copy_to_host()

        # remove temporary folder(s)
        self.tmp_files.extend([
            # self.stage_io_dict.get("unique_dir", ""),
            self.tmp_folder
        ])
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def pcz_zip(input_pdb_path: str, input_crd_path: str,
            output_pcz_path: str,
            properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`PCZzip <flexserv.pcasuite.PCZzip>`flexserv.pcasuite.PCZzip class and
    execute :meth:`launch() <flexserv.pcasuite.PCZzip.launch>` method"""

    return PCZzip(input_pdb_path=input_pdb_path,
                  input_crd_path=input_crd_path,
                  output_pcz_path=output_pcz_path,
                  properties=properties).launch()


def main():
    parser = argparse.ArgumentParser(description='Compress Molecular Dynamics (MD) trajectories using Principal Component Analysis (PCA) algorithms.', formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
    parser.add_argument('--config', required=False, help='Configuration file')

    # Specific args
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('--input_pdb_path', required=True, help='Input PDB file. Accepted formats: pdb.')
    required_args.add_argument('--input_crd_path', required=True, help='Input trajectory file. Accepted formats: crd, mdcrd, inpcrd.')
    required_args.add_argument('--output_pcz_path', required=True, help='Output compressed trajectory file. Accepted formats: pcz.')

    args = parser.parse_args()
    args.config = args.config or "{}"
    properties = settings.ConfReader(config=args.config).get_prop_dic()

    # Specific call
    pcz_zip(input_pdb_path=args.input_pdb_path,
            input_crd_path=args.input_crd_path,
            output_pcz_path=args.output_pcz_path,
            properties=properties)


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="biobb_flexserv",
    version="5.0.0",
    author="Biobb developers",
    author_email="adam.hospital@irbbarcelona.org",
    description="biobb_flexserv is a BioBB category for biomolecular flexibility studies on protein 3D structures.",
    long_description="biobb_flexserv allows the generation of protein conformational ensembles from 3D structures and the analysis of its molecular flexibility.",
    long_description_content_type="text/markdown",
    keywords="Bioinformatics Workflows BioExcel Compatibility Flexibility Ensembles Protein Structure",
    url="https://github.com/bioexcel/biobb_flexserv",
    project_urls={
        "Documentation": "http://biobb-flexserv.readthedocs.io/en/latest/",
        "Bioexcel": "https://bioexcel.eu/"
    },
    packages=setuptools.find_packages(exclude=['docs', 'test']),
    package_data={'biobb_flexserv': ['py.typed']},
    include_package_data=True,
    install_requires=['biobb_common==5.0.0'],
    python_requires='>=3.9',
    entry_points={
        "console_scripts": [
            "bd_run = biobb_flexserv.flexserv.bd_run:main",
            "dmd_run = biobb_flexserv.flexserv.dmd_run:main",
            "nma_run = biobb_flexserv.flexserv.nma_run:main",
            "pcz_zip = biobb_flexserv.pcasuite.pcz_zip:main",
            "pcz_unzip = biobb_flexserv.pcasuite.pcz_unzip:main",
            "pcz_animate = biobb_flexserv.pcasuite.pcz_animate:main",
            "pcz_bfactor = biobb_flexserv.pcasuite.pcz_bfactor:main",
            "pcz_collectivity = biobb_flexserv.pcasuite.pcz_collectivity:main",
            "pcz_evecs = biobb_flexserv.pcasuite.pcz_evecs:main",
            "pcz_hinges = biobb_flexserv.pcasuite.pcz_hinges:main",
            "pcz_info = biobb_flexserv.pcasuite.pcz_info:main",
            "pcz_lindemann = biobb_flexserv.pcasuite.pcz_lindemann:main",
            "pcz_stiffness = biobb_flexserv.pcasuite.pcz_stiffness:main",
            "pcz_similarity = biobb_flexserv.pcasuite.pcz_similarity:main"
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: Unix"
    ],
)



from . import pcz_zip
from . import pcz_unzip
from . import pcz_animate
from . import pcz_bfactor
from . import pcz_hinges
from . import pcz_stiffness
from . import pcz_similarity
from . import pcz_collectivity
from . import pcz_info
from . import pcz_evecs
from . import pcz_lindemann

name = "pcasuite"
__all__ = ["pcz_zip", "pcz_unzip", "pcz_animate", "pcz_bfactor", "pcz_hinges", "pcz_stiffness", "pcz_similarity", "pcz_collectivity", "pcz_info", "pcz_evecs", "pcz_lindemann"]
